@@ -202,12 +202,12 @@ if [ "$myhost" = "find-version" ]; then
 
     case "$version_parts" in
       2)
-        matchstr1="$package-$major.[0-9][0-9]*"
+        matchstr1="$package-$major.[0-9]*"
         matchstr2="$package-$major-base"
         vers="$major.$minor"
         ;;
       *)
-        matchstr1="$package-$major.$minor.[0-9][0-9]*"
+        matchstr1="$package-$major.$minor.[0-9]*"
         matchstr2="$package-$major.$minor-base"
         vers="$major.$minor.$micro"
         ;;
@@ -310,7 +310,7 @@ if [ "$myhost" = "w32" ]; then
 
     $tsdir/configure --enable-maintainer-mode ${SILENT} \
              --prefix=${w32root}  \
-             --host=${host} --build=${build} \
+             --host=${host} --build=${build} SYSROOT=${w32root} \
              ${configure_opts} ${extraoptions} "$@"
     rc=$?
     exit $rc
@@ -422,7 +422,7 @@ if [ -d .git ]; then
     [ -z "${SILENT}" ] && cat <<EOF
 *** Activating trailing whitespace git pre-commit hook. ***
     For more information see this thread:
-      http://mail.gnome.org/archives/desktop-devel-list/2009-May/msg00084html
+      https://mail.gnome.org/archives/desktop-devel-list/2009-May/msg00084.html
     To deactivate this pre-commit hook again move .git/hooks/pre-commit
     and .git/hooks/pre-commit.sample out of the way.
 EOF
